#!/bin/bash
########################################################################
# build_gcc.sh
#
# Build gcc and package for UPS.
#
# Adapted from build.sh.template provided by ssibuildshims v2_10_00.
########################################################################

########################################################################
# Non-customize-able boilerplate: skip to "Customize-able functions."
########################################################################

####################################
# Useful variables.
prog=${BASH_SOURCE[0]##*/}

##################
# Local (dumb) ssi_die() function for use until we source ssi_functions.
ssi_die() {
  printf "FATAL ERROR: ${*}\n" 1>&2; exit 1
}
##################

##################
# Basic usage function for use by ensure_buildshims until we have enough
# information to provide something more specific. Do NOT customize!
build_script_usage() {
    cat <<EOF
USAGE: ${prog} <ssi-msg-config-options> [--] <product-dir> <qualifier-args> <maketar>
       ${prog} -h|--help [<product-dir>]

  Build ${package} and package for UPS.

OPTIONS

  -h|--help

    Help. Supply <product-dir> for specific, detailed information on
    invoking ${prog}.

  <ssi-msg-config-options>

    Supply <product-dir> for more information.

ARGUMENTS

  <product-dir>

    The top-level UPS directory in which the source code distribution
    was unpacked. When specified with -h|--help, this enables us to
    provide more specific help about invoking ${prog}.

  <qualifier-args>
  <maketar>

    Supply <product-dir> for more information.

EOF
}
##################

##################
# Make sure we have UPS and ssibuildshims set up.
ensure_ssibuildshims() {
  if ! (( $# )) || [ "${@: -1}" = "-h" ] || [ "${@: -1}" = "--help" ]; then
    build_script_usage
    exit 1
  fi
  # Find <product-dir> from our arguments:
  local arg
  for arg in "${@}"; do
    [[ "$arg" ==  -* ]] && continue;
    product_dir="${arg}"; break
  done
  if [ -z "${product_dir}" ]; then
    # Bad usage.
    ssi_die "Bad <product-dir> ${product_dir}\n$(build_script_usage)"
    exit 2
  elif ! { [ -d "${product_dir}/.upsfiles" ] && [ -w "${product_dir}" ]; }; then
    # Something wrong with <product-dir>.
    ssi_die "<product-dir> not a valid, writable unified UPS directory: " \
      "${product_dir}\n$(build_script_usage)"
  fi
  # Promising...
  local abs_product_dir="$(cd "${product_dir}" 2>/dev/null 2>&1 && pwd -P)"
  (( $? )) && ssi_die "unable to cd to <product-dir>: ${product_dir}\n" \
    "$(build_script_usage)"
  # Find our own bootstraps and yank 'em.
  local ssibuildshims_topdir="${product_dir}/ssibuildshims/${ssibuildshims_version}"
  [ -f "${ssibuildshims_topdir}/bin/ssi_functions" ] && \
    source "${ssibuildshims_topdir}/bin/ssi_functions" >/dev/null 2>&1 && \
    ensure_ups "${product_dir}" && \
    setup ssibuildshims ${ssibuildshims_version} -z "${product_dir}" || \
    ssi_die "unable to find ssibuildshims ${ssibuildshims_version} in
${product_dir} and set it up with UPS."
  # Absolute is safer.
  absolute_path -- "${product_dir}" product_dir
}
##################

########################################################################
# Customize-able functions.
########################################################################

##################
# Ascertain whether we built and installed correctly. Add more /
# different calls to verify_in_filesystem as appropriate. Arguments
# (e.g. -q) are passed through to verify_in_filesystem.
is_built() { # *C*
  ${trace} verify_in_filesystem -f "${@}" "${pkgdir}/lib"{64,}"/libstdc++.${shext}" && \
    ${trace} verify_in_filesystem -f "${@}" "${pkgdir}/lib"{64,}"/libgfortran.${shext}" && \
    ${trace} verify_in_filesystem -x "${@}" "${pkgdir}/bin/gcc" && \
    ${trace} verify_in_filesystem -x "${@}" "${pkgdir}/bin/g++" && \
    ${trace} verify_in_filesystem -x "${@}" "${pkgdir}/bin/gfortran" && \
    ${trace} verify_in_filesystem -d "${@}" "${pkgdir}/include/c++"
}
##################

##################
# If a distribution archive is wanted and the installation is successful
# according to is_built(), then make it. Any arguments (e.g. -q) are
# passed through both to is_built() and make_basic_distribution().
#
# If make_basic_distribution() finds that the product is declared
# correctly and attempts to make the distribution, then it will cause us
# to exit with the status code of that operation. If not, it returns an
# error code.
#
# See make_basic_distribution -h for useful options to add to
# mbd_opts(). *C*
maybe_make_distribution() {
  if [ "${maketar}" = "tar" ]; then
    # Customize mbd_opts. *C*
    local mbd_opts=()
    ${trace} is_built "$@" && \
      ${trace} make_basic_distribution "${@}" "${mbd_opts[@]}"
  fi
}
##################

########################################################################
# Package-specific functions.
########################################################################

# Expand and patch sources.
unpack_and_patch() {
  # Failed globs should be empty: save the shell state and set it to
  # what we want.
  ssi_autosave_shell_state; shopt -s nullglob

  local source_stamp="${sourcedir}/sources.stamp"
  local patch_stamp="${sourcedir}/patches.stamp"
  local pkgs=(gcc) pkgtar pkg

  # Unwind the GCC source first.
  local gcctar="${tardir}/${pkgtarfile}"

  ${trace} prepare_source "${ssi_args[@]}" \
    -d "${sourcedir}" "${gcctar}" || exit

  # Now unwind all the dependency sources---no patching yet!
  for pkgtar in "${tardir}"/*.tar*; do
    [ "${pkgtar}" == "${gcctar}" ] && continue;
    pkg=${pkgtar##*/}
    pkg=${pkg%%-*}
    pkgs+=("${pkg}")
    ${trace} prepare_source "${ssi_args[@]}" \
      -d "${gcc_topdir}/${pkg}" \
      -e "--strip-components,1" "${pkgtar}" || exit
  done

  # Remember the current time.
  touch "$source_stamp"

  # Now check to see if we have a patch rollup archive newer than any of our
  # patches that we can use to bypass the patch application and the possible
  # need for regeneration of configure scripts. This became necessary on
  # 2020-03-30 when the mpfr-4.0.2 rollup patch as dynamically downloaded by
  # bootstrap.sh was updated to touch files triggering a regeneration of the
  # configure scripts. We must be very careful with dates to avoid re-triggering
  # the regeneration step on a machine that might not have an autotools suite of
  # sufficiently recent vintage.
  if [ -f "${rollup}" ] && \
       (( $(find "${patchdir}" -name '*.patch' -newer "${rollup}" \
                 2>/dev/null | wc -l) == 0 )); then
    ${trace} prepare_source "${ssi_args[@]}" -P "${patchdir}" \
      -d "${sourcedir}" "${rollup}" || \
      exit
    touch "${patch_stamp}"
    return
  fi

  # OK, we need to apply the patches and then the patch rollup archive must be
  # generated out-of-band using the ${patchdir}/make-patch-rollup script
  # according to ${patchdir}/00README.txt.
  local pad pcount=0
  for pkg in "${pkgs[@]}"; do
    pad=${pkg#gcc}
    ${trace} patch_source "${ssi_args[@]}" -P "${patchdir}" \
      -\# $(( pcount++ )) -a "${gcc_topdir}/${pad}" \
      "${patchdir}/${pkg}-"*.patch || \
      exit
  done
  touch "${patch_stamp}"

  # Stop with error for out-of-band rollup archive generation.
  ssi_die -i "patch rollup archive ${rollup} must be (re-)generated.\n" \
        "See patch/00README.txt"
}

########################################################################
# Main.
########################################################################

####################################
# Useful variables. *C*
thisdir=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd -P)
trace=maybe_xtrace # Could set to xtrace instead for unconditional.
(( ssi_trace = 1 )) # Trace certain commands by default.
####################################

####################################
# Package name and version information. *C*
. ${thisdir}/version_info.sh || \
  ssi_die "unable to load version info from ${thisdir}/version_info.sh"
####################################

##################
# Other useful variables. *C*
pkgtarname=${package}-${origpkgdotver}
pkgtarfile=${pkgtarname}.tar.gz
##################

####################################
# Handle options and arguments.

# Set up ssibuildshims and UPS / product_dir as early as possible.
ensure_ssibuildshims "${@}"

##################
# Now parse options and arguments, setting basequal, extraqual, maketar,
# etc.
#
# Use verify_quals() to parse command line options, verify that the user
# has provided qualifiers supported by this build script, and set
# expected variables such as basequal, extraqual, cqual, etc.
#
# Customize the extra usage text as appropriate. Note that DEBUG_PATCHES
# and VERBOSE_PATCHES are honored by prepare_source() and
# patch_source(). DEBUG_(CONFIG|BUILD|TEST|INSTALL) should be accounted
# for (or not) by the relevant command in this script.
#
# See verify_quals -h for options and examples.
verify_quals --parse --usage-file=- -- "$@" <<EOF || exit # *C*
ENVIRONMENT

  DEBUG_(CONFIG|BUILD|TEST|INSTALL)

    Verbose output from the corresponding step.

  DEBUG_PATCHES
  VERBOSE_PATCHES

    Command tracing and increased verbosity for source unpacking and patching
    operations. DEBUG_PATCHES will cause exit after patches have been applied.

  NO_PROFILE_BOOTSTRAP

    Disable the building of a profiled bootstrap to save build time at
    the expense of compile speed.
EOF
##################

##################
# Define globals: pkgdir, sourcedir, blddir, fullqual, etc. See
# define_basics -h for steering options.
${trace} define_basics TRACK_PROGRESS NO_QUAL NO_DEBUG_SOURCE || exit # *C*
##################

##################
# Variables requiring define_basics() to have been called.
gcc_topdir="${sourcedir}/${pkgtarname}"
rollup="${patchdir}/patch-rollup-${pkgdotver}.tar.bz2"
lddvar=LD_LIBRARY_PATH
if [ "${OS1}" = Darwin ]; then
  shext=dylib
  lddvar="DY${lddvar}"
  cc=/usr/bin/clang
  cxx=/usr/bin/clang++
else
  shext=so
fi
##################

####################################
# Declare ourselves privately, and then set ourselves up to allow setup
# of dependencies. *C*
#
# Comment or delete if not required:
# ${trace} fake_declare_product "${ssi_args[@]}" -x -- || exit
####################################

####################################
# Build-only dependencies. *C*
#
ssi_info "building bootstrap and dependencies with GCC $(gcc -dumpversion)"
####################################

# Print active products.
echo
ups active
echo

# Possible short-circuit if we're already built and declared.
${trace} maybe_make_distribution --silent

ssi_info "building $package for $OS${plat:+-$plat}${qualdir:+-$qualdir} (flavor $pkgflvr)"

# Ensure directories: pkgdir, sourcedir, blddir.
${trace} init_pkg_dirs || exit

####################################
# Prepare source for building. *C*

${trace} cd "${sourcedir}" \
  || ssi_die "unable to cd to source directory ${sourcedir}"

# Source tarballs and patches. See prepare_source -h / patch_source -h
# for options.
unpack_and_patch

# Bail early if DEBUG_PATCHES is set.
[ -n "${DEBUG_PATCHES}" ] && exit 1
####################################

####################################
# Prepare to build. *C*
#
# Set configure / compile flags, environment variables, etc.

##################
# Be especially careful with system PATH-like variables.

eval "export ${lddvar}=\"$pkgdir/lib64:$pkgdir/lib${GCC_FQ_DIR:+:$GCC_FQ_DIR/lib64:$GCC_FQ_DIR/lib}\""
export COMPILER_PATH="${pkgdir}/bin"
export PATH=$(dropit -sfe "${COMPILER_PATH}")

echo
for var in PATH ${lddpath} COMPILER_PATH; do
  eval "echo \"$var=\${${var}}\""
done
echo
##################

##################
# Configure parallelism. *C*
ncores=$(ncores)
##################

##################
# Configuration options.
languages="c,c++,fortran,lto"
conf_options=(
  ${cc:+CC=$cc}
  ${cxx:+CXX=$cxx}
  --disable-libgcj
  --enable-__cxa_atexit
  --enable-checking=release
  --enable-compressed-debug-sections=all
  --enable-languages=${languages}
  --enable-libstdcxx-time=rt
  --enable-plugin
  --enable-stage1-checking=all
  --enable-threads=posix
  --prefix=${pkgdir}
  --with-system-zlib
)

if [[ "$flvr" != "Linux64bit+2.6-2.12" ]]; then # not SLF6
  conf_options+=(--enable-lto)
fi
if [ "${OS1}" = Darwin ]; then
  sdk=$(xcrun --sdk macosx --show-sdk-path)
  [ -n "${sdk}" ] || ssi_die "unable to locate MacOS SDK"
  # Most headers are in the SDK now.
  conf_options+=(--with-native-system-header-dir=/usr/include \
    --with-sysroot="${sdk}" )
  # Also, no support for 32-bit from Catalina on.
  [[ "${OS}" =~ ^d([0-9]+) ]] && (( ${BASH_REMATCH[1]} >= 18 )) && \
    conf_options+=(--disable-multilib)
else
  # Go not supported on Darwin targets.
  languages="${languages+${languages},}go"
  # Don't require Zstd on Darwin.
  conf_options+=(--with-zstd)
fi
####################################

####################################
# Configure, build, and install.
#
# Use ${blddir}/${pkgtarname} for in-source builds, ${blddir} otherwise.

${trace} cd "${blddir}" || \
  ssi_die "unable to cd to build directory ${blddir}"

##################
# Configure. *C*
${trace} ${gcc_topdir}/configure "${conf_options[@]}" || \
  ssi_die "configure failed"
##################

##################
# Build. *C*

# The profiledbootstrap target is reputed to produce a GCC which
# compiles code around 7% faster according to some benchmarks, but the
# build of GCC itself could take significantly longer on MacOS (>17h on
# a modern High Sierra laptop).
if [ -z "${NO_PROFILE_BOOTSTRAP}" ]; then
  target=profiledbootstrap
else
  target=bootstrap
fi

# Compose the build command.
build_cmd=(make ${target} -j "${ncores}")
[ "${OS1}" = Darwin ] && \
  build_cmd+=("BOOT_LDFLAGS=-Wl,-headerpad_max_install_names")

# Execute the build.
${trace} "${build_cmd[@]}" || ssi_die "build failed"
##################

##################
# Test. *C*
##################

##################
# Install. *C*
${trace} make install || ssi_die "install failed"
##################

ssi_info "finished building ${package} ${pkgver}"

####################################
# Pre-declaration processing. *C*

${trace} rm_libtool_la || exit

ssi_info "making links for cc, c++ and f95"
cd "${pkgdir}/bin" || ssi_die "unable to cd to ${pkgdir}/bin"
{ [ -x cc ] || ${trace} ln -s ${ssi_verbose:+-v} gcc cc; } && \
  { [ -x c++ ] || ${trace} ln -s ${ssi_verbose:+-v} g++ c++; } && \
  { [ -x f95 ] || ${trace} ln -s ${ssi_verbose:+-v} gfortran f95; } || \
  ssi_die "unable to make links for cc, c++ and f95"

ssi_info "fixing path references in Python GDB files"
${trace} find "${pkgdir}/lib"* -name '*.py' | \
  ${trace} xargs perl -wpi'' -e \
  "s&(['\"])"'\Q'"${pkgdir}"'\E\b(.*?\g{1})&'\
'${1}\{prefix\}${2}.format(prefix=os.environ["GCC_FQ_DIR"])&g;' || \
  ssi_die "unable to fix GDB .py files."
####################################

##################
# Declare and set up the installed package. *C*
#
# Add -c for current declaration if required.
${trace} declare_product -x -- || exit
##################

#################
# Verify installation and make distribution archive. *C*
#
# To customize verification, see is_built(), since it is also used by
# maybe_make_distribution(). See the documentation for those functions
# (above) for customization opportunities.

# Add a message to ssi_build_failed if appropriate.
${trace} is_built && ${trace} ssi_build_completed && \
  ssi_info "${package} is installed at ${pkgdir}" && \
  ${trace} maybe_make_distribution || \
  ssi_build_failed # *C*
########################################################################
# Nothing beyond this point.
########################################################################
